import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import javax.microedition.rms.*;
import java.io.*;
import java.util.*;

public class Dictionary extends MIDlet implements CommandListener  
{ 
    TextBox mainMenu;
    static Command QUIT_CMD = new Command(Locale.EXIT, Command.EXIT, 2);
    static Command BACK_CMD = new Command(Locale.BACK, Command.BACK, 2);
    static Command TRANSLATE_CMD = new Command(Locale.TRANSLATE, Command.OK, 1);
    static Command ORDER_BY_FREQUENCY_CMD = new Command(Locale.BY_FREQUENCY, Command.ITEM, 1);
    static Command ORDER_BY_TIME_CMD = new Command(Locale.BY_TIME, Command.ITEM, 1);
    
    static final int MAX_LENGTH = 32;
    static final int MAX_FILE_SIZE = 32*1024;
    static final int SELECTION_LIMIT = 32;
    static final char SEPARATOR = ':';

    class Volume { 
        String first;
        String last;
    }

    Volume[] volumes;
    int      nVolumes;
    byte[]   buf = new byte[MAX_FILE_SIZE];
    int      prevVolumeSize;
    int      prevVolumeIndex;
    Vector   translatedWords = new Vector();
    RecordStore store;

    private void openStore() {
        try { 
            store = RecordStore.openRecordStore("DICTIONARY", true);
            int n = store.getNumRecords();
            translatedWords.setSize(n);
            for (int i = 0; i < n; i++) {
                byte[] record = store.getRecord(i+1);                    
                TranslatedWord tw = new TranslatedWord();
                tw.frequency = record[0] & 0xFF;
                tw.word = new String(record, 1, record.length-1);
                translatedWords.setElementAt(tw, i);
            }
        } catch(RecordStoreException x) { 
            x.printStackTrace();
            store = null;
        }
    }

    private void closeStore() 
    {
        if (store != null) { 
            try { 
                store.closeRecordStore();
                store = null;
            } catch(RecordStoreException x) { 
                x.printStackTrace();
            }
        }
    }

    void addTranslatedWord(String word) { 
        for (int i = 0, n = translatedWords.size(); i < n; i++) {
            TranslatedWord tw = (TranslatedWord)translatedWords.elementAt(i);
            if (tw.word.equals(word)) {
                if (tw.frequency < 0xFF) {
                    tw.frequency += 1;
                    if (store != null) {
                        byte[] record = tw.pack();
                        try { 
                            store.setRecord(i+1, record, 0, record.length);
                        } catch(RecordStoreException x) { 
                            x.printStackTrace();
                        }
                    }
                }
                return;
            }
        }
        TranslatedWord tw = new TranslatedWord();
        tw.word = word;
        tw.frequency = 1;
        if (store != null) {
            byte[] record = tw.pack();
            try { 
                store.addRecord(record, 0, record.length);
            } catch(RecordStoreException x) { 
                x.printStackTrace();
            }
        }
        translatedWords.addElement(tw);
    }
        
    public Dictionary ()
    {
        mainMenu = new TextBox(Locale.PREFIX, "", MAX_LENGTH, TextField.ANY);
        mainMenu.addCommand(QUIT_CMD);
        mainMenu.addCommand(TRANSLATE_CMD);
        buildIndex();
        openStore();
    }


    protected void destroyApp(boolean unconditional) 
    {
        mainMenu = null;
        closeStore();
    }    

    protected  void pauseApp() 
    {
    }

    protected void startApp() 
    {
        Display.getDisplay(this).setCurrent(mainMenu);
        mainMenu.setCommandListener(this) ;
    }
/**********************
    static String getString(byte[] buf, int start, int len) { 
        /*
        char[] chars = new char[len];
        for (int i = 0; i < len; i++) { 
            byte b = buf[start+i];
            chars[i] = (char)(b < 0 ? b == -88 ? 0x401 : b == -72 ? 0x451 : 
                              (0x400 | ((b & 0x7f) - 0x30)) : b);
        }
        return new String(chars);
        */
        /*
        char[] chars = new char[len];
        for (int i = 0; i < len; i++) { 
            byte b = buf[start + i];            
            chars[i] = (char)(buf[start + i] & 0xFF);
        }
        return new String(chars);        
    }

    static int getUnicodeStringLength(byte[] buf, int start) { 
        int n;
        while ((n = buf[start]) != 0) { 
            start += n + 2;
        }
        return start + 1;
    }
            
    static String getUnicodeString(byte[] buf, int start) { 
        char[] chars = new char[256];
        int i = start, j = 0, n; 
        while ((n = buf[i++] & 0xFF) != 0) {
            int charset = buf[i++] << 8;
            while (--n >= 0) { 
                chars[j++] = (char)(charset | (buf[i++] & 0xFF));
            }
        }
        return new String(chars, 0, j);
    }
***************/
    static String rusLowerCase(String str)
    {
        char[] chars = new char[str.length()];
        str.getChars(0, chars.length, chars, 0);
        for (int i = 0; i < chars.length; i++) { 
            char ch = chars[i];
            if ((ch >= 'A' && ch <= 'Z') || (ch >= 0x410 && ch <= 0x42F)) { 
                chars[i] += 0x20;
            }
        }
        return new String(chars);
    }

    static String getString(byte[] buf, int start, int len) {         
        char[] chars = new char[len];
        for (int i = 0; i < len; i++) { 
            byte b = buf[start+i];
            chars[i] = (char)(b < 0 ? b == -88 ? 0x401 : b == -72 ? 0x451 : 
                              (0x400 | ((b & 0x7f) - 0x30)) : b);
        }
        return new String(chars);
    }

    static int getUnicodeStringLength(byte[] buf, int start) { 
        while (buf[start++] != '\n');
        return start;
    }
            
    static String getUnicodeString(byte[] buf, int start) { 
        int i = start;
        while (buf[i] != '\n') { 
            i += 1;
        }
        return getString(buf, start, i-start);
    }


    private void buildIndex()
    {
        int size;
        try 
        { 
            InputStream in = getClass().getResourceAsStream("/dictionary.idx");
            size = in.read(buf);
            in.close();
        } 
        catch (IOException x) 
        { 
            throw new Error("Failed to read index");
        }
        Vector volumesList = new Vector();
        for (int i = 0; i < size; i++) 
        { 
            int j; 
            Volume v = new Volume();
            for (j = i; buf[j] != ':'; j++);
            v.first = getString(buf, i, j-i);
            for (i = ++j; buf[i] != '\n'; i++);
            v.last = getString(buf, j, i-j);
            volumesList.addElement(v);
        }
        nVolumes = volumesList.size();
        volumes = new Volume[nVolumes];
        prevVolumeIndex = -1;
        for (int i = 0; i < nVolumes; i++) 
        { 
            volumes[i] = (Volume)volumesList.elementAt(i);
        }
    }

    public Vector getSelection(String prefix) 
    { 
        if (nVolumes == 0) { 
            buildIndex();
        }
        Vector selection = new Vector();
        int i;
        int n = nVolumes; 
        Volume[] vs = volumes;
        boolean found = false;
        try { 
            for (i = 0; i < n && vs[i].last.compareTo(prefix) < 0; i++);
            while (i < n) { 
                int size;
                if (prevVolumeIndex != i) {                     
                    InputStream in = getClass().getResourceAsStream("volume." + (i+1));
                    size = in.read(buf);
                    prevVolumeIndex = i;
                    prevVolumeSize = size;
                    in.close();
                } else { 
                    size = prevVolumeSize;
                }
                String prevWord = vs[i].first;
                for (int j = 0; j < size;) { 
                    int start = j;
                    while (buf[j] != SEPARATOR) { 
                        j += 1;
                    }
                    String word = (buf[start] >= '0' && buf[start] <= '9') 
                        ? prevWord.substring(0, buf[start]-'0') + getString(buf, start+1, j-start-1)
                        : getString(buf, start, j-start);
                    start = ++j;
                    j = getUnicodeStringLength(buf, j);
                    if (word.compareTo(prefix) >= 0) { 
                        if (word.startsWith(prefix)) { 
                            found = true;
                        } else if (found) { 
                            return selection;
                        }
                        selection.addElement(new Translation(word, getUnicodeString(buf, start)));
                        if (selection.size() == SELECTION_LIMIT) { 
                            return selection;
                        }
                    }
                    prevWord = word;
                }
                i += 1;
            }
	} catch (IOException x)   { 
	    x.printStackTrace();
	}
        return selection;
    }
    
    

    public void commandAction(Command c, Displayable d) 
    {
        if (c == QUIT_CMD) 
        {
            destroyApp(true);
            notifyDestroyed();
        } 
        else 
        {
            String key = rusLowerCase(mainMenu.getString());
            if (key.length() == 0) {
                if (translatedWords.size() > 0) { 
                    new TranslatedList(this);
                }
            } else {
                mainMenu.setTitle(Locale.ONE_MINUTE);
                mainMenu.setString(Locale.SEARCHING);
                Vector selection = getSelection(key);
                switch (selection.size()) { 
                case 0:
                    Alert a = new Alert(key, Locale.NO_VARIANTS, null, AlertType.ERROR);
                    a.setTimeout(Alert.FOREVER);
                    Display.getDisplay(this).setCurrent(a, mainMenu);
                    break;
                case 1:
                    Translation t = (Translation)selection.firstElement();
                    if (t.word.equals(key)) { 
                        new TranslationBox(this, mainMenu, (Translation)selection.firstElement());
                        break;
                    }
                default:
                    new PrefixList(this, mainMenu, selection);
                    break;
                }
                mainMenu.setTitle(Locale.PREFIX);
                mainMenu.setString(key);
            }
        }
    }
}



